/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.bridge;

import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashSet;
import java.util.Iterator;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.dbms.AbstractConnectionFactory;
import org.xquark.mapper.mapping.MappingIterator;
import org.xquark.mapper.mapping.RepositoryMapping;
import org.xquark.mapper.storage.MapperTupleBuilder;
import org.xquark.mapper.storage.SAXHandler;
import org.xquark.mapper.storage.ValidatingFilerAdapter;
import org.xquark.schema.SchemaManager;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.xml.xdbc.XMLDBCException;

/**
 * An object implementing this class represents a processed mapping file.
 * Objects of this class are thread-safe and can be reused multiple times.<BR>
 * However, as they encapsulate a JDBC connection, it is better
 * to use different objects for different threads.<BR>
 * This class uses a cache for loaded mapping files, so that the loading overhead is
 * reduced. The cache mechanism can be overriden to reload a modified file.
 */

public class Mapping
{
	private static final String 	RCSRevision 	= "$Revision: 1.5 $";
	private static final String 	RCSName     	= "$Name:  $";
    
    private AbstractConnection connection = null;
    private String systemId = null;
    private RepositoryMapping metadata = null;
    private HashSet mappers = new HashSet();
    private SchemaManager schemaManager;  // this is to be independant if MappingMetadata is reset
    
    /**
     * Constructor.
     * @param conn a JDBC connection
     * @param source the SAX InputSource encapsulating the mapping file
     * @param reload true means that the mapping file should be reloaded, even if it is present in the cache
     * @param loader ClassLoader used for loading user generators
     * @throws IOException if an IO error occured while accessing the mapping file
     * @throws SAXException if a SAX error occured while parsing the mapping file
     * @throws SQLException if an SQL error occured while accessing the database
     */
    public Mapping(Connection conn, InputSource source, boolean reload, ClassLoader loader)
    throws IOException, SAXException, SQLException
    {
        connection = AbstractConnectionFactory.newConnection(conn, true);
        systemId = source.getSystemId();
        metadata = MappingMetadata.loadMapping(connection, source, reload, loader);
        schemaManager = MappingMetadata.getSchemaManager();
    }
    
	/**
	 * Constructor.
	 * @param conn a JDBC connection
	 * @param source the SAX InputSource encapsulating the mapping file
	 * @param reload true means that the mapping file should be reloaded, even if it is present in the cache
	 * @throws IOException if an IO error occured while accessing the mapping file
	 * @throws SAXException if a SAX error occured while parsing the mapping file
	 * @throws SQLException if an SQL error occured while accessing the database
	 */
	public Mapping(Connection conn, InputSource source, boolean reload)
	throws IOException, SAXException, SQLException
	{
		this(conn, source, reload, Mapping.class.getClassLoader());
	}
    
    /**
     * Constructor.
     * @param conn a JDBC connection
     * @param fileURI the URI of the mapping file.
     * @param reload true means that the mapping file should be reloaded, even if it is present in the cache
     * @param loader ClassLoader used for loading user generators
     * @throws IOException if an IO error occured while accessing the mapping file
     * @throws SAXException if a SAX error occured while parsing the mapping file
     * @throws SQLException if an SQL error occured while accessing the database
     */
    public Mapping(Connection conn, String fileURI, boolean reload, ClassLoader loader)
    throws IOException, SAXException, SQLException
    {
        this(conn, new InputSource(fileURI), reload, loader);
    }
    
	/**
	 * Constructor.
	 * @param conn a JDBC connection
	 * @param fileURI the URI of the mapping file.
	 * @param reload true means that the mapping file should be reloaded, even if it is present in the cache
	 * @throws IOException if an IO error occured while accessing the mapping file
	 * @throws SAXException if a SAX error occured while parsing the mapping file
	 * @throws SQLException if an SQL error occured while accessing the database
	 */
	public Mapping(Connection conn, String fileURI, boolean reload)
	throws IOException, SAXException, SQLException
	{
		this(conn, new InputSource(fileURI), reload);
	}
    
    /**
     * Frees all resources associated to the object, and removes it from the cache
     * @throws XMLDBCException if a database connection error occured.
     */
    public synchronized void close() throws XMLDBCException, SQLException
    {
        Iterator it = mappers.iterator();
        while (it.hasNext()) ((Mapper)it.next()).close();
    }
    
    /**
     * Creates a new Mapper object for storing a set of XML documents
     * in the database
     * @return a Mapper object
     * @throws XMLDBCException if a database connection error occured.
     * @see Mapper
     */
    public Mapper getMapper() throws XMLDBCException
    {
        SchemaValidationContext validationContext = new SchemaValidationContext(schemaManager);
        ValidatingFilerAdapter validator = new ValidatingFilerAdapter(validationContext, schemaManager);
        SAXHandler handler = new SAXHandler
        (
        new MappingIterator(metadata),
        connection,
        new MapperTupleBuilder
        (
        connection,
        metadata,
        validationContext
        )
        );
        validator.setSAXHandler(handler);
        Mapper result = new MapperImpl(this, validator, handler, handler, handler);
        mappers.add(result);
        return result;
    }
    
    /**
     * Returns the JDBC2 connection used by this object to access the database.
     * @return a JDBC2 connection
     */
    public Connection getConnection()
    { return connection.getConnection(); }
    
    /**
     * Returns the system id (the URL) of the mapping file, if known.
     * @return the system id of the mapping file, or null if unknown.
     */
    public String getSystemId()
    { return systemId; }
    
    /**
     * Package-private method returning the internal AbstractConnection
     * @return the internal AbstractConnection used by this object.
     */
    AbstractConnection getAbstractConnection()
    { return connection; }
    
    /**
     * Package-private method returning the internal description of the
     * mapping file
     * @return the internal representation of the mapping file
     */
    RepositoryMapping getMetaData()
    { return metadata; }
    
    /**
     * Package-private maethod used to remove a Mapper from its parent list
     */
    void remove(MapperImpl mapper)
    {
        mappers.remove(mapper);
    }
}


