/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery.parser;

import java.util.*;

import org.xquark.schema.SchemaManager;
import org.xquark.util.NamespaceContextStack;
import org.xquark.xquery.metadata.StaticContext;
import org.xquark.xquery.parser.normalize.ExpressionNormalizationVisitor;
import org.xquark.xquery.parser.util.ArrayListUtil;
import org.xquark.xquery.parser.util.Constants;

/*
[30]    Module    ::=    VersionDecl? (MainModule | LibraryModule) 
[31]    MainModule    ::=    Prolog QueryBody 
[32]    LibraryModule    ::=    ModuleDecl Prolog 
[33]    ModuleDecl    ::=    <"module" "namespace"> NCName "=" StringLiteral Separator 
[34]    Prolog    ::=    ((NamespaceDecl
| XMLSpaceDecl
| DefaultNamespaceDecl
| DefaultCollationDecl
| BaseURIDecl
| SchemaImport
| ModuleImport
| VarDecl
| ValidationDecl
| FunctionDecl) Separator)* 
[35]    Separator    ::=    ";" 
*/

public class XQueryModule implements Cloneable {
	private static final String RCSRevision = "$Revision: 1.25 $";
	private static final String RCSName = "$Name:  $";
	
	protected String version = null;
	protected String prefix = null;
	protected String namespace = null;
	protected NamespaceContextStack declarations = null;
	protected String defaultFunctionNamespace = null;
	protected int xmlspace = Constants.NOTHING;
	protected String defaultCollation = null;
	protected String baseURI = null;
	
	protected HashMap hashVarList = null;
	protected ArrayList varList = null;
	protected HashMap hashExtVarList = null;	
	protected HashMap hashFuncList = null;
	protected ArrayList funcList = null;
	protected HashMap importSchemas = null;
	protected ArrayList schemaList = null;
	protected HashMap importModules = null;
	protected ArrayList moduleList = null;
	
//	protected int lastVarIndex = 0;
	
	protected int validation = Constants.NOTHING;
	protected ArrayList expressions = null;
	
	protected SchemaManager schemaManager = null;
	protected StaticContext staticContext = null;

	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(ParserVisitor visitor) throws XQueryException {
		visitor.visit(this);
	}

	// #############################################################################
	// CONSTRUCTOR STUFF
	// #############################################################################

	public XQueryModule() {}

	// #############################################################################
	// ATTRIBUTE ACCESS STUFF
	// #############################################################################

	// expressions
	public ArrayList getExpressions() {
		return expressions;
	}
	public void setExpressions(ArrayList expressions) throws XQueryException {
		this.expressions = expressions;
		if (expressions != null)
			for (int i = 0; i < this.expressions.size(); i++) {
				((XQueryExpression) this.expressions.get(i)).setParentModule(this);
				((XQueryExpression) this.expressions.get(i)).setParentExpressions(null);
			}
	}

	public void setElementAt(XQueryExpression expr, int index) {
		expr.setParentModule(this);
		expressions.set(index, expr);
	}

	// declarations
	public NamespaceContextStack getDeclarations() {
		return declarations;
	}
	public void setDeclarations(NamespaceContextStack declarations) {
		this.declarations = declarations;
	}

	public String getNameSpace(String prefixname) {
		if (prefixname == null)
			return null;
		if (declarations == null)
			return null;
		String str = declarations.getNamespaceURI(prefixname);
		if (str == null)
			return null;
		int len = str.length();
		if (str.indexOf("\"") == 0)
			return str.substring(1, len - 2);
		else
			return str;
	}

	// variables
	public HashMap getVariables() {
		return hashVarList;
	}
	public void setVariableList(ArrayList varList) {
		this.varList = varList;
	}
	public ArrayList getVariableList() {
		return varList;
	}
	public void setVariables(HashMap hashVarList) {
		this.hashVarList = hashVarList;
		if (hashVarList != null) {
			Collection varValues = hashVarList.values();
			Iterator itValues = varValues.iterator();
			while (itValues.hasNext()) {
					HashMap map = (HashMap)itValues.next();
					Collection mapValues = map.values();
					Iterator itMapValues = mapValues.iterator();
					while (itMapValues.hasNext()) {
						((Variable) itMapValues.next()).setParentModule(this);
					}
			}
		}
	}

	public Variable getVariableDefinition(String name) {
		if (name == null || hashVarList == null)
			return null;
		return (Variable) hashVarList.get(name);
	}
	
	// external variables
	public HashMap getExternalVariables() {
		return hashExtVarList;
	}
	public void setExternalVariables(HashMap hashExtVarList) {
		this.hashExtVarList = hashExtVarList;
	}
	
	// functions
	public HashMap getFunctions() {
		return hashFuncList;
	}
	public ArrayList getFunctionList() {
		return funcList;
	}
	public void setFunctionList(ArrayList funcList) {
		this.funcList = funcList;
	}
	public void setFunctions(HashMap hashFuncList) {
		this.hashFuncList = hashFuncList;
		if (hashFuncList != null) {
			Collection funcValues = hashFuncList.values();
			Iterator itValues = funcValues.iterator();
			while (itValues.hasNext()) {
					HashMap map = (HashMap)itValues.next();
					Collection mapValues = map.values();
					Iterator itMapValues = mapValues.iterator();
					while (itMapValues.hasNext()) {
				 		((FunctionDeclaration) itMapValues.next()).setParentModule(this);
					}
			}
		}
	}

//	public FunctionDeclaration getFunctionDefinition(String name) {
//		if (name == null || hashFuncList == null)
//			return null;
//		return (FunctionDeclaration) hashFuncList.get(name);
//	}
	
	// default collation
	public void setDefaultCollation(String defaultCollation) {
		this.defaultCollation = defaultCollation;
	}

	public String getDefaultCollation() {
		return defaultCollation;
	}
	
	// xquery version
	public void setVersion(String version) {
		this.version = version;
	}

	public String getVersion() {
		return version;
	}

	// module prefix
	public void setPrefix(String prefix) {
		this.prefix = prefix;
	}

	public String getPrefix() {
		return prefix;
	}

	// module location
	public void setNamespace(String location) {
		this.namespace = location;
	}

	public String getNamespace() {
		return namespace;
	}

	// default function namespace
	public void setDefaultFunctionNamespace(String namespace) {
		this.defaultFunctionNamespace = namespace;
	}

	public String getDefaultFunctionNamespace() {
		return defaultFunctionNamespace;
	}

	// base uri
	public void setBaseURI(String uri) {
		this.baseURI = uri;
	}

	public String getBaseURI() {
		return baseURI;
	}

	// xmlspace
	public void setXmlSpace(int xmlspace) {
		this.xmlspace = xmlspace;
	}

	public int getXmlSpace() {
		return xmlspace;
	}

	// validation
	public void setValidation(int validation) {
		this.validation = validation;
	}

	public int getValidation() {
		return validation;
	}

	// import schemas
	public HashMap getImportSchemas() {
		return importSchemas;
	}
	public ArrayList getSchemaList() {
		return schemaList;
	}
	public void setSchemaList(ArrayList schemaList) {
		this.schemaList = schemaList;
	}
	public void setImportSchemas(HashMap importSchemas) {
		this.importSchemas = importSchemas;
		if (importSchemas != null) {
			Collection importSchemaValues = importSchemas.values();
			Iterator itValues = importSchemaValues.iterator();
			while (itValues.hasNext())
				 ((ImportSchema) itValues.next()).setParentModule(this);
		}
	}

	public ImportSchema getImportSchema(String name) {
		if (name == null || importSchemas == null)
			return null;
		return (ImportSchema) importSchemas.get(name);
	}

	// import modules
	public HashMap getImportModules() {
		return importModules;
	}
	public void setModuleList(ArrayList moduleList) {
		this.moduleList = moduleList;
	}
	public ArrayList getModuleList() {
		return moduleList;
	}
	public void setImportModules(HashMap importModules) {
		this.importModules = importModules;
		if (importModules != null) {
			Collection importModuleValues = importModules.values();
			Iterator itValues = importModuleValues.iterator();
			while (itValues.hasNext())
				 ((ImportModule) itValues.next()).setParentModule(this);
		}
	}

	public ImportModule getImportModule(String name) {
		if (name == null || importModules == null)
			return null;
		return (ImportModule) importModules.get(name);
	}

	// schemaManager
	public void setSchemaManager(SchemaManager schemaManager) {
		this.schemaManager = schemaManager;
	}
	public SchemaManager getSchemaManager() {
		return schemaManager;
	}

	// staticContext
	public void setStaticContext(StaticContext staticContext) {
		this.staticContext = staticContext;
	}
	public StaticContext getStaticContext() {
		return staticContext;
	}

//	// varIndex
//	public void setLastVarIndex(int index) {
//		lastVarIndex = index;
//	}
//	
//	public int getLastVarIndex() {
//		return lastVarIndex;
//	}
	
	// #############################################################################
	// NORMALIZE STUFF
	// #############################################################################

	public void normalize(TypeVisitor typeVisitor) throws XQueryException {
		normalize(typeVisitor, false);
	}

	public void normalize(TypeVisitor typeVisitor, boolean isMediator) throws XQueryException {
		try {
			ExpressionNormalizationVisitor normalizeVisitor = new ExpressionNormalizationVisitor(typeVisitor);
			accept(normalizeVisitor);
		} catch (XQueryException ne) {
			if (namespace != null)
				throw new XQueryException(ne.getMessage()+" in module "+namespace);
			else
				throw new XQueryException(ne.getMessage()+" in main module");
		}

		try {
			if (expressions != null && !expressions.isEmpty()) {
				RestructureVisitor restructureVisitor = new RestructureVisitor(typeVisitor, isMediator);
				XQueryExpression expr = null;
				for (int i = 0; i < expressions.size(); i++) {
					XQueryExpression expri = (XQueryExpression)expressions.get(i);
//					if (expri instanceof XQueryExpressionSequence) {
//						this.expressions.remove(i);
//						this.expressions.addAll(i,((XQueryExpressionSequence)expri).getSubExpressions());
//						i--;
//					} else {
						expri.accept(restructureVisitor);
						if (restructureVisitor.getResult() != null)
							expr = restructureVisitor.getResult();
						if (expr != null) {
							expressions.set(i, expr);
						}
//					}
				}
			}
		} catch (XQueryException xqe) {
			if (namespace != null)
				throw new XQueryException(xqe.getMessage()+" in module "+namespace);
			else
				throw new XQueryException(xqe.getMessage()+" in main module");
		}
	}

	// #############################################################################
	// SOURCES STUFF
	// #############################################################################

	public ArrayList getSourceNames() {
		ArrayList sources = null;
		if (expressions != null) {
			sources = new ArrayList();
			for (int i = 0; i < expressions.size(); i++) {
				ArrayListUtil.add(sources, ((XQueryExpression) expressions.get(i)).getSourceNames());
			}
		}
		return sources;
	}

	public ArrayList getUrls() {
		ArrayList urls = null;
		if (expressions != null) {
			urls = new ArrayList();
			for (int i = 0; i < expressions.size(); i++) {
				ArrayListUtil.add(urls, ((XQueryExpression) expressions.get(i)).getUrls());
			}
		}
		return urls;
	}

	public String toString(boolean lineFeeds, boolean imports) {
		PrintVisitor pv = new PrintVisitor();
		pv.reset(lineFeeds, imports);
		try {
			this.accept(pv);
		} catch (XQueryException qe) {}
		return pv.getString();
	}

	public String toString() {
		PrintVisitor pv = new PrintVisitor();
		pv.reset(false, true);
		try {
			this.accept(pv);
		} catch (XQueryException qe) {}
		return pv.getString();
	}
}
